/** 
* ezrep.sp : Sourcemod Reporting Helper Plugin for TF2Center
* http://naknak.tf/code/ezrep/
*
* by naknak, http://steamcommunity.com/id/_naknak_/ 
*
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License, version 3.0, as published by the
* Free Software Foundation. See <http://www.gnu.org/licenses/>. 
* As a special exception, you have permission to use this program and/or its
* derivative works under the same terms as SourceMod itself.  In particular, 
* you may link this code to Valve's game engines. SourceMod's license provides
* further exceptions to the GPL, see <http://www.sourcemod.net/license.php>.
*
*/
#define PLUGIN_VERSION "1.01.20140526"
public Plugin:myinfo =
{
	name = "ezrep",
	author = "naknak",
	description = "Easy menu-driven reporting for TF2Center lobbies. Type !r to start",
	version = PLUGIN_VERSION,
	url = "http://naknak.tf/"
};

#include <regex>
// used for guessing defaults, the menu will always offer both choices
#define MIN6S_PLAYERS 9  
#define MAX6S_PLAYERS 12

// for repping purposes, HL/4v4/Ulti are all the same
#define NUM_GAMEMODES 2
#define NUM_CLASSES 9
#define NUM_TEAMS 2

// when incrementing UNDEF via a specifier like "2", increment this class instead 
#define DEFAULT 0 

// invalid index to use for undef values
#define UNDEF 9 

// in seconds
#define MENU_TIMEOUT 12

// universal string size, in bytes
// TODO: only the tokenizer really needs this much space, most else will fit in 32
#define SZ 256
#define SS SZ-1

#define DEBUG 0
#define DPRINT DEBUG && PrintToChat
#define DLOG DEBUG && PrintToServer

new String:GameTypes[][SZ] = { "HL/4v4", "6v6" };
new String:TeamLabels[][SZ] = { "BLU", "RED", "", "", "", "", "", "", "", "" };
new const String:ClassLabels[][][SZ] = { 
	{ "Scout", "Soldier", "Pyro", "Demoman", "Heavy", "Engineer", "Medic", "Sniper", "Spy", "" },
	{ "Scout 1", "Scout 2",  "Roamer", "Pocket", "Demoman", "Medic", "", "", "", "" }
};
new const String:Targets[][][SZ] = { 
	{ "scout", "solly", "pyro", "demo", "heavy", "engie", "med", "sniper", "spy" },
	{ "scout 1", "scout 2", "roamer", "pocket", "demo", "med", "", "", "" }
};
// inter-gamemode mappings
// when class is specified, and team is not, and the team selection thru menu changes the gamemode, class may change too
// hl scout/solly are invalid in 6s (not specific enough)
// 6s scouts/soldiers map to hl scout/solly.  HL/6s demo/med are equivalent.
// hl pyro ("p") can mean pocket in 6s
// all other mappings are invalid
#define SCOUT 0
#define POCKET 2
#define SOLLY 1
#define DEMO6S 4
#define DEMOHL 3
#define MED6S 5
#define MEDHL 6
new const Equivalence[][] = {
	{ UNDEF, UNDEF, POCKET, DEMO6S, UNDEF, UNDEF, MED6S, UNDEF, UNDEF, UNDEF },
	{ SCOUT, SCOUT, SOLLY, SOLLY, DEMOHL, MEDHL, UNDEF, UNDEF, UNDEF, UNDEF }
}

new Handle:reTokenizer;
new Handle:reValidator;
new Handle:reClassTokens[NUM_GAMEMODES][UNDEF+1];
new Handle:reTeamTokens[UNDEF+1];
new Handle:reIncrTokens[NUM_GAMEMODES][UNDEF+1];
new Handle:reNone;
new Handle:reLaunch;
new Handle:reClose;

// new Handle:cvarDebug;
new Handle:cvarAlways;
new Handle:cvarLaunchAnnounce;
new Handle:cvarLaunchAnnounceDelay;
new Handle:cvarLaunchAction;
new Handle:cvarLaunchActionDelay;
new Handle:cvarCloseAnnounce;
new Handle:cvarCloseAnnounceDelay;
new Handle:cvarCloseAction;
new Handle:cvarCloseActionDelay;

new bool:inLobby;
new gameType=UNDEF;

// unmatchable RE for invalid values.  match-everything for the catchall UNDEF value.
#define NONE "(?!.)."
#define ALL ""

// trigger regex activates the rep logic.  validation regex decides what messages can be ignored (because TF2C will recognize them)
#define RE_TRIGGER "^\\s* [/!]r[eport]{0,7}  "
#define RE_TOKEN "(?:[^a-z1-2]+([a-z1-2]+))"
#define RE_VALID "^\"?(?i: !rep(ort)? [ ] (red|blue?) [ ] (scout([ ][12])?|solly|soldier|(soldier_)?(roamer|pocket)|pyro|demo(man)?|heavy|engie|engineer|med(ic)?|sniper|spy) )\"?$"
#define RE_LAUNCH "^\"*GameType:[ ](\\S+)"
#define RE_CLOSE  "Lobby[ ]\\#\\d+[ ]closed:"

// setup actions, done once per map load
// complile regexes, create cvars
public OnPluginStart()
{
	// cvarDebug = CreateConVar("ezrep_debug", "0", "print lots of debugging info", _, true, 0.0, true, 1.0);
	cvarAlways = CreateConVar("ezrep_always", "0", 
		"run all the time.  If 0, plugin is activated by 'rcon say GameType: ___' and deactivated by 'rcon say Lobby #____ closed:'",
		_, true, 0.0, true, 1.0
		);
	cvarLaunchAnnounce = CreateConVar("ezrep_launch_announce", "^1This server has ezrep!  Type ^4!r^1 to report players! Easily!", 
		"send this message on lobby launch, after ezrep_launch_delay seconds have passed.  Use color codes: ^1 for normal, ^3 for light green, ^4 for green, ^5 for olive, ^6 for yellow"
		);
	cvarLaunchAnnounceDelay = CreateConVar("ezrep_launch_announce_delay", "300", 
		"delay ezrep_launch_announce by this many seconds",
		_, true, 1.0, true, 9999.0
		);
	cvarLaunchAction = CreateConVar("ezrep_launch_action", "mp_tournament_restart", 
		"run this command on lobby launch after ezrep_launch_action_delay seconds have passed"
		);
	cvarLaunchActionDelay = CreateConVar("ezrep_launch_action_delay", "10", 
		"delay ezrep_launch_action by this many seconds", 
		_, true, 1.0, true, 9999.0
		);
	cvarCloseAnnounce = CreateConVar("ezrep_close_announce", "^1 Lobby over! ^4 Get out!^1 ", 
		"send this message on lobby close, after ezrep_close_announce_delay seconds have passed. Use color codes: ^1 for normal, ^3 for light green, ^4 for green, ^5 for olive, ^6 for yellow"
		);
	cvarCloseAnnounceDelay = CreateConVar("ezrep_close_announce_delay", "10", 
		"delay ezrep_close_action by this many seconds", 
		_, true, 1.0, true, 9999.0
		);
	cvarCloseAction = CreateConVar("ezrep_close_action", "mp_tournament_restart", 
		"run this on lobby close after ezrep_close_action_delay seconds have passed"
		);
	cvarCloseActionDelay = CreateConVar("ezrep_close_action_delay", "120", 
		"delay ezrep_close_action by this many seconds", 
		_, true, 1.0, true, 9999.0
		);

	reNone = regexc(NONE);
	new const String:TeamTokens[][SZ]=	{ "^b", "^r(?!o)",  NONE, NONE, NONE, NONE, NONE, NONE, NONE, ALL };
	new const String:ClassTokens[][][SZ] = { 
		{ "^sc", "^so", "^p(?!oc)", "^d", "^h|^fat", "^e", "^m|^doc", "^sn", "^sh?p", ALL },
		{ "^sc[^2]*$", "^sc.*2", "^r(?!e)|^so[^2]*$", "^p(?!y)|^so.*2", "^d", "^m|^doc", NONE, NONE, NONE, ALL }
	};
	new const String:IncrTokens[][][SZ] = {
		{ NONE, NONE,  NONE,        NONE, NONE, NONE, NONE, NONE, NONE, NONE },
		{ "^2", NONE,  "^2|^p(?!y)", NONE, NONE, NONE, NONE, NONE, NONE, "^2" }
	}
	for (new i=0; i<=UNDEF; i++) {
		reTeamTokens[i] = regexc(TeamTokens[i]);
		for (new j=0; j<NUM_GAMEMODES; j++) {
			reClassTokens[j][i] = regexc(ClassTokens[j][i]);
			// DLOG("about to compile /%s/ [j=%d i=%d]", IncrTokens[j][i],j,i);
			reIncrTokens[j][i] = regexc(IncrTokens[j][i]);
		}
	}
	new String:Tokenizer[SZ];
	Format(Tokenizer, SS, "%s %s? %s? %s? %s? %s?", RE_TRIGGER, RE_TOKEN, RE_TOKEN, RE_TOKEN, RE_TOKEN, RE_TOKEN);
	reTokenizer = regexc(Tokenizer);
	reValidator = regexc(RE_VALID);
	reLaunch = regexc(RE_LAUNCH);
	reClose = regexc(RE_CLOSE);

	inLobby = GetConVarBool(cvarAlways);
}

// callback for user making a menu choice: update the selector and go back to Complete_Report to see if anything's still missing
public Handle_Menu(Handle:menu, MenuAction:action, client, arg)
{
	switch(action)
	{
		case MenuAction_Select:
		{
			new String:selector[SZ];
			GetMenuItem(menu, arg, selector, SS);
			new is6s=(selector[0]-'0') % 2;
			new team=(selector[1]-'0') % 10;
			new class=(selector[2]-'0') % 10;
			Complete_Report(client,is6s,team,class);
		}
 
		case MenuAction_End:
		{
			CloseHandle(menu);
		}
	}
	return 0;
}
 
static Handle:regexc(const String:regex[]) { 
	new String:error[SZ];
	new RegexError:errno;
	new Handle:re;
	if (reNone && StrEqual(regex, NONE)) { //memoize
		return reNone;
	} else { 
		// DLOG("       compiling /%s/", regex);
		re = CompileRegex(regex, PCRE_EXTENDED, error, SS, errno);
		if (!re)   PrintToServer("regex /%s/ gave invalid handle 0", regex);
		if (errno) PrintToServer("regex /%s/ gave error: %s", regex, error);
		return re;
	}
}
	


static isThis6s() 
{
	if (gameType != UNDEF) return gameType;
	new players=0;
	new maxplayers=GetMaxHumanPlayers();
	for (new i=1; i<=maxplayers; i++) {
		if (IsClientConnected(i) && IsClientAuthorized(i) && !IsClientReplay(i) && !IsClientSourceTV(i) &&!IsFakeClient(i)) players++
	}
	new guess=((players >= MIN6S_PLAYERS) && (players <= MAX6S_PLAYERS)) ? 1 : 0;
	DLOG("[ezrep:isThis6s] players=%d isThis6s=%d", players, guess);
	return guess;
}

// hook to the say* family of commands.  All reports start here
public Action:OnClientSayCommand(client, const String:cmd[], const String:arg[])
{
	new start=0;

	if(!client) {
		if (MatchRegex(reLaunch, arg)) { 
			new String:tokenmatch[SZ];
			GetRegexSubString(reLaunch, 1, tokenmatch, SS);
			gameType=StrEqual(tokenmatch, "6v6") ? 1:0; 
			inLobby=true;
			PrintToServer("[ezrep] launch detected");
			CreateTimer(GetConVarFloat(cvarLaunchAnnounceDelay), Announce, 1);
			CreateTimer(GetConVarFloat(cvarLaunchActionDelay), Act, 1);
		} else if (MatchRegex(reClose, arg)) { 
			PrintToServer("[ezrep] close detected");
			inLobby=GetConVarBool(cvarAlways);
			gameType=UNDEF;
			CreateTimer(GetConVarFloat(cvarCloseAnnounceDelay), Announce, 0);
			CreateTimer(GetConVarFloat(cvarCloseActionDelay), Act, 0);
		}
		return Plugin_Continue;
	}

	if(arg[start]=='"') start++; // random " abound
	if (arg[start]=='!'  && CharToLower(arg[start+1])=='r' && MatchRegex(reValidator, arg)) { 
		PrintToChat(client, "[ezrep] sent %s", arg);
		return Plugin_Continue;
	}

	if (!inLobby) return Plugin_Continue;

	DPRINT(client, "[ezrep:OnClientSayCommand] you said [%s] (%c) %d chars", arg, arg[0], strlen(arg));
	while ((start < SS) && arg[start]==' ') start++;
	if ( arg[start]=='!' || arg[start]=='/' && CharToLower(arg[start+1])=='r') {
		new String:said[SZ]="";
		StrCat(said, SS, arg[start]);
		for (new i=0; i<strlen(said); i++) said[i]=CharToLower(said[i]); 
		DPRINT(client, "[ezrep:OnClientSayCommand] I see [%s] %d chars", said, strlen(said));

		new is6s=isThis6s();
		new team=UNDEF;
		new class=UNDEF;

		new matched = MatchRegex(reTokenizer, said);
		DPRINT(client, "regex matched %d pieces", matched-1);

		for (new i=1; i<matched; i++)
		{
			new bool:consumed=false;
			new String:token[SZ];
			GetRegexSubString(reTokenizer, i, token, SS);
			DPRINT(client, "token %d: [%s]", i, token);
			for (new j=0; j<UNDEF; j++) !consumed && team==UNDEF  &&  MatchRegex(reTeamTokens[j], token) && (++consumed) && (team=j);
			for (new j=0; j<UNDEF; j++) !consumed && class==UNDEF &&  MatchRegex(reClassTokens[is6s][j], token) && (++consumed) && (class=j);
			if (!consumed) { 
				if (MatchRegex(reIncrTokens[is6s][class], token)) { 
					++consumed;
					class=1+( class==UNDEF ? DEFAULT : class);
				} else if (MatchRegex(reIncrTokens[is6s ^ 1][class], token)) { 
					++consumed;
					is6s ^= 1;
					new String:tokenmatch[SZ];
					GetRegexSubString(reIncrTokens[is6s ^ 1][class], 0, tokenmatch, SS);
					DPRINT(client, "gamemode changed to %d by presence of incr token [%s] matching [%s] for class %d", is6s, token, tokenmatch, class)
					class=1+( class==UNDEF ? DEFAULT : class);
				}
			}
			if (!consumed) DPRINT(client, "ignored token");
		}
		DPRINT(client, "[ezrep:OnClientSayCommand] is6s=%d, team=%d, class=%d", is6s,team,class);

		Complete_Report(client, is6s, team, class);
	}
	return Plugin_Continue;
}

// present a team menu (if we don't know the team) or a class menu (if we do know the team) or just send a report if we know all.
static Complete_Report(const client, const is6s, const team, const class)
{
	DPRINT(client, "[ezrep:Complete_Report] is6s=%d, team=%d, class=%d", is6s,team,class);
	if (team==UNDEF || class==UNDEF) { 
		new Handle:repmenu = CreateMenu(Handle_Menu);
		new String:title[SZ] = "";
		Format(title, SS, "!rep %s %s", 
			( (team==UNDEF && class==UNDEF || team!=UNDEF) ? TeamLabels[team] : "___"), 
			( (team==UNDEF && class==UNDEF || class!=UNDEF)  ? ClassLabels[is6s][class] : "___")
		);
		SetMenuTitle(repmenu, title);
		if (team==UNDEF) { 
			AddMenuItem(repmenu, "", "", ITEMDRAW_SPACER);
			for(new j=0; j<=1; j++) { 
				AddMenuItem(repmenu, "", "", ITEMDRAW_SPACER);
				for(new i=0; i<NUM_TEAMS; i++) { 
					new String:sel[SZ];
					DPRINT(client, "[ezrep:Complete_Report] menu item %d: is6s=%d, team=%d, class=%d", i, is6s,team,class);
					Format(sel, SS, "%d%d%d", is6s ^ j, i, (j ? Equivalence[is6s][class] : class) );

					new String:label[SZ] 
					Format(label, SS, "%s %s", TeamLabels[i], (j ? GameTypes[is6s ^ j] : "") );
					
					AddMenuItem(repmenu, sel, label);
					DPRINT(client, sel);
				}
			}
		} else if (class==UNDEF) { 
			for(new i=0; i<NUM_CLASSES; i++) { 
				if (strlen(Targets[is6s][i])) { 
					new String:sel[SZ];
					Format(sel, SS, "%d%d%d", is6s,team,i);
					AddMenuItem(repmenu, sel, ClassLabels[is6s][i]);
				}
			}
		}
		SetMenuPagination(repmenu, MENU_NO_PAGINATION);
		SetMenuExitButton(repmenu, true);
		return DisplayMenu(repmenu, client, MENU_TIMEOUT);
	} else {
		CreateTimer(1.05, Submit_Report, client*1000 + is6s*100 + team*10 + class)
	}
	return 0;
}

// a callback fired by short-lived event timer (to defeat the final !rep from being considered a flood)
// only get one int of data here, so it's packed into 5 digits: PPMTC
// PP = player id (1..MAXPLAYERS)
// M = mode (1=6v6, 0=everything else)
// T = team (BLU/RED = 0/1)
// C = class (0..8, same ordering TF2 uses)
public Action:Submit_Report(Handle:timer, any:selector)
{
	new String:report[SZ]="";
	new client = (selector / 1000) % (MAXPLAYERS+1);
	new is6s = ((selector % 1000) /100) % 2;
	new team = ((selector % 100) /10) % 2;
	new class = selector % 10;
	Format(report, SS, "say !rep %s %s", TeamLabels[team], Targets[is6s][class]);
	PrintToServer("[ezrep] faking on behalf of client %d: %s", client, report);
	FakeClientCommandEx(client, report);
}

public Action:Announce(Handle:timer, any:selector)
{
	new String:msg[SZ];
	GetConVarString( selector ? cvarLaunchAnnounce : cvarCloseAnnounce, msg, SS );
	PrintToServer("[ezrep] announcing [%s] inlobby=%d", msg, selector);
	// simple colorizer, tf2 uses control chars ^A..^F (ASCII 1..6) followed by mandatory space:
	// \x01 = Normal
	// \x02 = None
	// \x03 = Light green
	// \x04 = Green
	// \x05 = Olive
	// \x06 = Yellow/black 
	// translate all '^n' for n '1' .. '6' to ASCII control character n plus space
	new len=strlen(msg);
	for (new i=0; i<len-1; i++) {
		if (msg[i]=='^' && msg[i+1]>='1' &&  msg[i+1]<='6')  {
			msg[i]=msg[i+1]-'0';
			msg[i+1]=' '
		}
	}
	PrintToChatAll(msg);
}

public Action:Act(Handle:timer, any:selector)
{
	new String:msg[SZ];
	GetConVarString( selector ? cvarLaunchAction : cvarCloseAction, msg, SS );
	PrintToServer("[ezrep] executing [%s] inlobby=%d", msg, selector);
	ServerCommand(msg);
}
// EOF
